--[[--------------------   TODO   ------------------------------

- remove need for some translations: metatable that *GetItemInfo* on "name" access
	then have to change the "group" attribute to be automatique (#index)
- include weightstones for maces, add a flag for (weight|sharp)stones,
  and check weapon type to filter accordingly.
- better implementation for hideLowRanks
- may have to use GetItemCount on weaponID, so when there's 2 or more
	i retrieve poison by slot instead of weaponID
- alternative layout with count inside icon: GetItemCount >> frame:CreateFontString( nil, "ARTWORK", "NumberFontNormalSmall" )
- customizations: time coloration: whole text, time left only, none
- 3rd weapon
--------------------------------------------------------------]]

local header = CreateFrame( "Frame", nil, nil, "SecureStateHeaderTemplate" )
local menu = CreateFrame( "Frame", nil, nil, "ABRP_FrameTemplate" )

header:SetAttribute( "addchild", menu )
menu:SetAttribute( "hidestates", 0 )


AraPoisonsDBPC = {}
local config
local defaultConfig = {
	icon = "weapon",
	text = "poison",
	hideLowRanks = false,
	hideWeaponBuffFrame = false,
	notifyLosses = true,
}
local L = setmetatable( { }, { __index = function(t,k) return k end } )
local playerGUID, inCombat
local blank = { name=L["Undefined"] }
local dropdown = CreateFrame( "Frame", "ABRP_DDFrame", nil, "UIDropDownMenuTemplate" )

local ICON_SIZE, BUTTON_HEIGHT, GAP, TEXT_OFFSET = 20, 22, 9, 4

local next, format, GetSpellInfo, GetWeaponEnchantInfo, GetItemCount, GetItemIcon, GetItemInfo, GetInventorySlotInfo, GetInventoryItemLink =
	next, format, GetSpellInfo, GetWeaponEnchantInfo, GetItemCount, GetItemIcon, GetItemInfo, GetInventorySlotInfo, GetInventoryItemLink


-------------------[[  Poisons | Weightstones | oils  ]]-------------------

local poisons = {
	-- sharpening stones
	[ 2862] = { name=L["Rough Sharpening Stone"],		buffText=L["Sharpened (+2 Damage)"] },
	[ 2863] = { name=L["Coarse Sharpening Stone"],		buffText=L["Sharpened (+3 Damage)"] },
	[ 2871] = { name=L["Heavy Sharpening Stone"],		buffText=L["Sharpened (+4 Damage)"] },
	[ 7964] = { name=L["Solid Sharpening Stone"],		buffText=L["Sharpened (+6 Damage)"] },
	[12404] = { name=L["Dense Sharpening Stone"],		buffText=L["Sharpened (+8 Damage)"] },
	[18262] = { name=L["Elemental Sharpening Stone"],	buffText=L["+28 Critical Rating"] },
	[23122] = { name=L["Consecrated Sharpening Stone"],	buffText=L["+100 Attack Power vs Undead"] },
	[23528] = { name=L["Fel Sharpening Stone"],		buffText=L["Sharpened (+12 Damage)"] },
	[23529] = { name=L["Adamantite Sharpening Stone"],	buffText=L["Sharpened (+14 Crit Rating and +12 Damage)"] },
	-- oils
	[34539] = { name=L["Righteous Weapon Coating"] },
}

-- reverse lookup (in same table :p)
for poisonID, poisonInfo in next, poisons do
	if( type(poisonID)=="number" )then
		poisonInfo.id = poisonID
		poisons[poisonInfo.buffText or poisonInfo.name] = poisonInfo
	end
end

local romanNumbers = { "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X" }
local function SetPoison( name, ... )
	for i=1, select( "#", ... ) do
		local poisonInfo = { group = name, rank = i, name = format( "%s%s%s%s", name, i>1 and " " or "", i>1 and romanNumbers[i>=10 and 10 or i] or "", i>10 and romanNumbers[i-10] or "" ), id = select(i,...) }
		poisons[poisonInfo.id], poisons[poisonInfo.name] = poisonInfo, poisonInfo
	end
end
SetPoison( L["Crippling Poison"], 3775, 3776 )
SetPoison( L["Instant Poison"], 6947, 6949, 6950, 8926, 8927, 8928, 21927 )
SetPoison( L["Mind-numbing Poison"], 5237, 6951, 9186 )
SetPoison( L["Deadly Poison"], 2892, 2893, 8984, 8985, 20844, 22053, 22054 )
SetPoison( L["Wound Poison"], 10918, 10920, 10921, 10922, 22055 )
SetPoison( L["Anesthetic Poison"], 21835 )


-----------------------------[[  utilities  ]]-----------------------------

local function print( msg, ... )
	DEFAULT_CHAT_FRAME:AddMessage( select("#",...)>0 and format(msg,...) or msg )
end


local colorex = { { v=0, r=255,g=64,b=64 }, { v=1800, r=255,g=255,b=64 }, { v=3600, r=64,g=255,b=64 } }
local function GetHexColor( value )
	local upper, lower = colorex[1]
	for i=1, #colorex do
		lower, upper = upper, colorex[i]
		if( upper.v > value )then break end
		if( i == #colorex )then lower = upper end
	end
	if( lower == upper )then return format( "%.2x%.2x%.2x", lower.r, lower.g, lower.b ) end
	local percent = (value - lower.v) / (upper.v - lower.v)
	return format( "%.2x%.2x%.2x", (upper.r - lower.r) * percent + lower.r, (upper.g - lower.g) * percent + lower.g, (upper.b - lower.b) * percent + lower.b )
end


local function SecondsToTime( seconds )
	return format( "|cff%s%i:%.2i|r", GetHexColor(seconds), floor(seconds/60), seconds%60 )
end


local function SetButtonHighlight( button )
	local highlight = button:CreateTexture()-- nil, "BACKGROUND" )
	highlight:SetTexture( "Interface\\QuestFrame\\UI-QuestTitleHighlight" )
	highlight:SetAlpha( .7 )
--	highlight:SetAllPoints( button )
	highlight:SetPoint( "LEFT", button, "LEFT", ICON_SIZE, 0 )
	highlight:SetPoint( "RIGHT", button, "RIGHT" )
	button:SetHighlightTexture( highlight )
end


local function GetWeaponInfo( slot )
	local itemLink = GetInventoryItemLink( "player", GetInventorySlotInfo( slot ) ) or ""
	local name, _, _, _, _, _, type, _, _, icon = GetItemInfo( itemLink )
	local id = tonumber( strmatch( itemLink, "item:(%d+)" ) )
	return id, icon, name, type
end


local function SetWeaponBuffFrameVisibility()
	 local frame = WeaponBuffaloFrame or TemporaryEnchantFrame
	 if( config.hideWeaponBuffFrame )then
		frame:Hide()
		if( frame == TemporaryEnchantFrame )then BuffFrame:SetPoint( TemporaryEnchantFrame:GetPoint() ) end
	 else
		frame:Show()
	 end
end

---------------------------[[  secure buttons  ]]--------------------------


local function SecureMenu_OnUpdate( self, elapsed )
	menu.timer = menu.timer - elapsed
	if( menu.timer > 0 )then return end
	menu:SetScript( "OnUpdate", nil )
	header:SetAttribute( "state", 0 ) -- hide secure menu (menu + buttons)
end


-- SecureMenu_OnEnter/Leave are set to launchers, menu, and buttons
local function SecureMenu_OnEnter()
	menu:SetScript( "OnUpdate", nil )
	menu.timer = .2
end

local function SecureMenu_OnLeave()
	menu:SetScript( "OnUpdate", SecureMenu_OnUpdate )
end


local buttons = setmetatable( { }, { __index = function(t,i)
	local b = CreateFrame( "Button", nil, menu, "SecureActionButtonTemplate" )
	rawset( t, i, b )
--	header:SetAttribute( "addchild", b )
--	b:SetAttribute( "hidestates", 0 )
	b:SetAttribute( "type", "item" )
	b:SetHeight( BUTTON_HEIGHT )
	b:SetScript( "OnEnter", SecureMenu_OnEnter )
	b:SetScript( "OnLeave", SecureMenu_OnLeave )

	b.icon = b:CreateTexture()
	b.icon:SetWidth( ICON_SIZE ) b.icon:SetHeight( ICON_SIZE )
	b.icon:SetPoint( "LEFT", b, "LEFT" )

	b.font = b:CreateFontString( nil, "OVERLAY" )
	local font, _, flags = NumberFontNormal:GetFont()
	b.font:SetFont(font, 11, flags)
	b.font:SetJustifyH( "RIGHT" )
	b:SetFontString(b.font) -- allow proper "Set(Highlight)TextColor"
	SetButtonHighlight(b)
	b:SetPoint( "TOPLEFT", menu, "TOPLEFT", GAP, BUTTON_HEIGHT * (1-i) - GAP ) -- ADDED
	b.font:SetPoint( "LEFT", b, "LEFT", ICON_SIZE + TEXT_OFFSET, 0 )
	return b
end } )


local sortedPoisons = {}

local function SortPoisons()
	for i=1, #sortedPoisons do sortedPoisons[i] = nil end
	for poisonID, poisonInfo in next, poisons do
		if( type(poisonID) == "number" )then
			poisonInfo.count = GetItemCount( poisonID )
			if( poisonInfo.count > 0 )then
				if( not poisonInfo.icon )then poisonInfo.icon = GetItemIcon( poisonID ) end
				sortedPoisons[#sortedPoisons+1] = poisonInfo
			end
		end
	end
	sort( sortedPoisons, function(a,b) return (a.group and (a.group == b.group)) and (config.hideLowRanks and a.rank > b.rank or a.rank < b.rank) or a.name < b.name end )

	if( config.hideLowRanks )then
		local prevGroup
		for i = #sortedPoisons, 1, -1 do
			local poisonInfo = sortedPoisons[i]
			if( prevGroup and (prevGroup == poisonInfo.group) )then
				tremove( sortedPoisons, i )
			else
				prevGroup = poisonInfo.group
			end
		end
	end

	return sortedPoisons
end


local function Launcher_OnEnter( self ) -- self.block OK
	if( inCombat )then return end -- TODO: possibly use alpha to hide/show for combat use.
	SecureMenu_OnEnter()
	local menu_width, item_width = 0
	local lastPoison = config[self.block.weaponID] or config[self.block.slot] or blank
	for i, poisonInfo in ipairs( SortPoisons() ) do
		local button = buttons[i]
		button.font:SetFormattedText( "%s [|cffffffff%i|r]", poisonInfo.name, poisonInfo.count )
		button.icon:SetTexture( poisonInfo.icon )
		button:SetAttribute( "item", "item:" .. poisonInfo.id )
		button:SetAttribute( "target-item", "item:" .. (self.block.weaponID or 0) )
	--	button:SetAttribute( "target-slot", self.block.slot == "MainHandSlot" and 16 or 17 )
		if( lastPoison.id == poisonInfo.id )then
			button.font:SetTextColor( .9, 0.45, .1 )
		else
			button.font:SetTextColor( 1, 0.823529, 0 )
		end
		item_width = ICON_SIZE + TEXT_OFFSET + button.font:GetStringWidth()
		if( item_width > menu_width )then menu_width = item_width end
	end
	for i=1, #sortedPoisons do
		local button = buttons[i]
		button:SetWidth( menu_width )
	end
	menu:SetWidth( menu_width + GAP*2 )
	menu:SetHeight( BUTTON_HEIGHT * #sortedPoisons + GAP*2 )

	local showBelow = select( 2, self:GetCenter() ) > ( UIParent:GetHeight() / 2 )
	menu:ClearAllPoints()
	menu:SetPoint( showBelow and "TOP" or "BOTTOM", self, showBelow and "BOTTOM" or "TOP" )

	header:SetAttribute( "state", #sortedPoisons > 0 and 1 or 0) -- show menu
	-- show items that were hidden in a previous shorter list
	for i=1, #sortedPoisons do buttons[i]:Show() end
	-- hide items that were shown in a previous longer list
	for i=#sortedPoisons+1, #buttons do buttons[i]:Hide() end
end


local dropdown_init
local function Launcher_ShowMenu( self, button )
	header:SetAttribute( "state", 0 )
	UIDropDownMenu_Initialize( dropdown, dropdown_init, "MENU" )
	ToggleDropDownMenu( 1, nil, dropdown, self, 0, 0)
end


local function CreateLauncher()
	local launcher = CreateFrame( "Button", nil, UIParent, "SecureActionButtonTemplate,SecureAnchorEnterTemplate" )

	launcher:RegisterForClicks( "AnyUp" )
	launcher:SetAttribute( "type1", "item" )
	launcher:SetAttribute( "type2", "menu" )
	launcher.menu = Launcher_ShowMenu
	launcher:SetScript( "OnEnter", Launcher_OnEnter )
	launcher:SetScript( "OnLeave", SecureMenu_OnLeave )
	launcher:SetScript( "OnMouseUp", function(self, button)
		if( button == "MiddleButton" )then
			CancelItemTempEnchantment( self.block.slot == "MainHandSlot" and 1 or 2 )
		end
	end )
	return launcher
end

local mainButton = CreateLauncher()
local offButton = CreateLauncher()


----------------------------[[  data feeders  ]]---------------------------

-- would be nice for displays to provide some user-args or a ref to the DO
-- to tell apart which DO it is when more than one use the same OnEnter
function Block_OnEnter( displayFrame, launcherButton )
	if( not inCombat )then
		launcherButton.display = displayFrame
		launcherButton:RegisterForDrag( "LeftButton" )
		launcherButton:SetScript( "OnDragStart", function(self)
	--	launcherButton:SetScript( "OnMouseDown", function(self,button)
	--		if(button=="LeftButton")then
			self.display:GetScript( "OnDragStart" )( self.display )
	--		end
		end )
		launcherButton:SetScript( "OnDragStop", function(self)
	--	launcherButton:SetScript( "OnMouseUp", function(self,button)
	--		if(button=="LeftButton")then
			self.display:GetScript( "OnDragStop" )( self.display )
	--		end
		end )
		launcherButton:SetAllPoints( displayFrame )
		launcherButton:Show()
		launcherButton.block.OnEnter = Launcher_OnEnter
	end
end


local blockMH = LibStub("LibDataBroker-1.1"):NewDataObject( "|cFFFFB366Ara|r Poisons - Main Hand", {
	text = "Main Hand",
	icon = select(3, GetSpellInfo(2842) ),
	OnEnter = function(self) Block_OnEnter( self, mainButton ) end,
	slot = "MainHandSlot"
} )
mainButton.block = blockMH


local blockOH = LibStub("LibDataBroker-1.1"):NewDataObject( "|cFFFFB366Ara|r Poisons - Off Hand", {
	text = "Off-Hand",
	icon = select(3, GetSpellInfo(2842) ),
	OnEnter = function(self) Block_OnEnter( self, offButton ) end,
	slot = "SecondaryHandSlot"
} )
offButton.block = blockOH

local blocks = { blockMH, blockOH }


---------------------[[  events (and related) stuff  ]]---------------------


function SetLauncherAction( block, weaponID, poisonID )
	if( block and not inCombat )then -- TODO: better way to get the right one
		local launcherButton = blockMH == block and mainButton or offButton
		if( not poisonID )then poisonID = config[weaponID] or config[block.slot] or 0 end
		launcherButton:SetAttribute( "item", "item:" .. ( type(poisonID) == "table" and poisonID.id or poisonID ) )
		launcherButton:SetAttribute( "target-item", "item:" .. weaponID )
	end
end


local function UpdateWeaponInfo()
	for _, block in next, blocks do
		local id, icon, name, type = GetWeaponInfo( block.slot )
		if( block.id ~= id )then
			block.weaponID, block.weaponIcon, block.weaponName, block.weaponType = id, icon, name, type
			SetLauncherAction( block, id )
		end
	end
end


function AddWeaponPoisonPair( weaponID, poisonInfo, newlyApplied )
--	print( "POISON APPLIED/REMOVED! Saving couple %s-%s.", weaponID, poisonInfo.id )
	local block = weaponID == blockMH.weaponID and blockMH or blockOH -- 3rd weapon inc
	if( not poisonInfo.icon )then poisonInfo.icon = GetItemIcon( poisonInfo.id ) end
	config[weaponID] = poisonInfo
	config[block.slot] = poisonInfo

	SetLauncherAction( block, weaponID, poisonInfo.id )
	if( newlyApplied )then
		-- wait for onupdate to update display
	elseif( config.notifyLosses )then -- notify lost poisons
		print( "|cffff4040-- |T%s:17|t %s |T%s:17|t --|r", GetItemIcon(poisonInfo.id), poisonInfo.buffText or poisonInfo.name, GetItemIcon(weaponID) )
	end
end


-- may have to split it in one OnUpdate per DO
function OnUpdate( self, elapsed ) -- self = header
	self.timer = self.timer - elapsed
	if( self.timer > 0 )then return end
	self.timer = 1
	_, blocks[1].timeLeft, _, _, blocks[2].timeLeft = GetWeaponEnchantInfo()

	for _, block in next, blocks do
		local poisonInfo = config[block.slot] or blank
		block.text = format( " %s%s",
			config.text == "poison" and poisonInfo.name or config.text == "weapon" and block.weaponName or "",
			format( config.text == "none" and "%s" or " [%s]",
				block.timeLeft and SecondsToTime( block.timeLeft/1000 ) or "|cffff4040n/a|r" ) )
		if( config.icon == "poison" and not poisonInfo.icon )then poisonInfo.icon = GetItemIcon(poisonInfo.id ) end
		block.icon = config.icon == "poison" and poisonInfo.icon or config.icon == "weapon" and block.weaponIcon or ""
	end
end


function OnEvent( self, event, ... ) -- self = header
	if( event == "PLAYER_REGEN_DISABLED" )then -- enter combat
		inCombat = true
		blockMH.prevWeaponID, blockOH.prevWeaponID = blockMH.weaponID, blockOH.weaponID
		header:SetAttribute( "state", 0 )
	elseif( event == "PLAYER_REGEN_ENABLED" )then -- leave combat
		inCombat = false
		if( blockMH.weaponID ~= blockMH.prevWeaponID )then SetLauncherAction( blockMH, blockMH.weaponID ) end
		if( blockOH.weaponID ~= blockOH.prevWeaponID )then SetLauncherAction( blockOH, blockOH.weaponID ) end
	elseif( event == "UNIT_INVENTORY_CHANGED" )then
		if( (...) ~= "player" )then return end
		UpdateWeaponInfo() -- TODO: bucket with OnUpdate
	elseif( event == "VARIABLES_LOADED" )then
		config = AraPoisonsDBPC
		for k,v in pairs(defaultConfig) do
			if( config[k] == nil )then config[k] = v end
		end

		SetWeaponBuffFrameVisibility()
		for _, block in next, blocks do
			local weaponID, weaponIcon, slot = GetWeaponInfo( block.slot )
			if( weaponID and ( config[weaponID] or config[ block.slot ] ) )then
				SetLauncherAction( block, weaponID, config[weaponID] or config[ block.slot ] )
			end
		end
	elseif( event == "PLAYER_LOGIN" )then
		-- TODO: possibly merge VARIABLES_LOADED and PLAYER_LOGIN
		playerGUID = UnitGUID( "player" )
		UpdateWeaponInfo()
		self:RegisterEvent( "UNIT_INVENTORY_CHANGED" )
	end
end


function OnCombatEvent( frame, event, timestamp, combatEvent, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, enchantName, weaponID, weaponName )
	if( dstGUID ~= playerGUID )then return end
	if( combatEvent ~= "ENCHANT_APPLIED" and combatEvent ~= "ENCHANT_REMOVED" )then return end
	local poisonInfo = poisons[enchantName]
	if( not poisonInfo )then return end
	AddWeaponPoisonPair( weaponID, poisonInfo, combatEvent == "ENCHANT_APPLIED" )
end


header.timer = 1
header:SetScript( "OnEvent", OnEvent )
header:SetScript( "OnUpdate", OnUpdate )
header:RegisterEvent( "PLAYER_REGEN_DISABLED" )
header:RegisterEvent( "PLAYER_REGEN_ENABLED" )
header:RegisterEvent( "VARIABLES_LOADED" )
if( IsLoggedIn() )then OnEvent( header, "PLAYER_LOGIN" ) else header:RegisterEvent( "PLAYER_LOGIN" ) end

menu:SetScript( "OnEvent", OnCombatEvent )
menu:RegisterEvent( "COMBAT_LOG_EVENT_UNFILTERED" )

menu:EnableMouse(true)
menu:SetScript( "OnEnter", SecureMenu_OnEnter )
menu:SetScript( "OnLeave", SecureMenu_OnLeave )


local options = {
{	text = "|cFFFFB366Ara|r Poisons (r3)",
	isTitle = true, },
{
	text = L["Hide low ranks"],
	desc = "If you have, say, \"Deadly Poison II\" and \"Deadly Poison IV\", the menu will only show \"Deadly Poison IV\".",
	get = function() return config.hideLowRanks end,
	set = function() config.hideLowRanks = not config.hideLowRanks end,
	sticky = true, },
{
	text = L["Hide weapon buff frame"],
	get = function() return config.hideWeaponBuffFrame end,
	set = function()
		config.hideWeaponBuffFrame = not config.hideWeaponBuffFrame
		SetWeaponBuffFrameVisibility()
	end,
	sticky = true, },
{
	text = L["Notify losses (in chat)"],
	get = function() return config.notifyLosses end,
	set = function() config.notifyLosses = not config.notifyLosses end,
	sticky = true, },
{
	text = format( "|cffaaaaff%s|r", L["Remove MH poison"] ),
	exec = function() CancelItemTempEnchantment(1) end,
	},
{
	text = L["Set icon as"],
	submenu = {
	{	text = L["Set icon as"],
		isTitle = true, },
	{	text = "Weapon",
		get = function() return config.icon == "weapon" end,
		set = function() config.icon = "weapon" end, },

	{	text = "Poison",
		get = function() return config.icon == "poison" end,
		set = function() config.icon = "poison" end, },

	{	text = "None",
		get = function() return config.icon == "none" end,
		set = function() config.icon = "none" end, }, }, },
{
	text = L["Set text as"],
	submenu = {
	{	text = L["Set text as"],
		isTitle = true, },
	{	text = "Weapon",
		get = function() return config.text == "weapon" end,
		set = function() config.text = "weapon" end, },

	{	text = "Poison",
		get = function() return config.text == "poison" end,
		set = function() config.text = "poison" end, },

	{	text = "None",
		get = function() return config.text == "none" end,
		set = function() config.text = "none" end, }, }, },
}


dropdown_init = function(level)
	level = level or 1
	for i, v in ipairs( level > 1 and UIDROPDOWNMENU_MENU_VALUE or options ) do
		local info = UIDropDownMenu_CreateInfo()
		info.text = v.text
		info.isTitle = v.isTitle
		info.tooltipTitle = v.desc and v.text
		info.tooltipText = v.desc
		info.checked = v.get
		info.keepShownOnClick = v.sticky
		info.func = v.set or v.exec
		info.hasArrow = v.submenu ~= nil
		info.value = v.submenu
		UIDropDownMenu_AddButton( info, level )
	end
end